import 'dart:async';
import 'dart:convert';

import 'package:flutter/widgets.dart';
import 'package:flutterx_http/src/x_http_client.dart';
import 'package:flutterx_utils/flutterx_utils.dart';
import 'package:http/http.dart';

/// Perform the call to the [client] with given [params]
typedef RequestFactory = Future<Response> Function(XHttpClient client, Map<String, dynamic> params);

/// Converts the [Response] into [T]
typedef ResponseConverter<T> = FutureOr<T> Function(Response body);

/// [ApiCall] represents an HTTP Call using [XHttpClient]
/// A call has a [request] factory and a response [convert]er
///
/// [T] the return type of the call (It would be helpful if it's an instance of [DTO])
///
/// ```dart
/// class MyApi {
///   final XHttpClient _client;
///
///   MyApi(String baseUrl) : _client = XHttpClient(baseUrl: baseUrl);
///
///   static final ApiCall<MyData> _getMyData = ApiCall.object<MyData>(
///       request: (client, params) => client.get('/api/myData', params: params), fromJson: MyData.fromJson);
///
///   Future<MyData> getMyData(String userId) => _getMyData(_client, {'userId': userId});
/// }
///
/// class MyData with DTO {
///   final String name;
///   final int age;
///
///   const MyData({required this.name, required this.age});
///
///   static MyData fromJson(JsonObject json) => MyData(name: json['name'], age: json['age']);
///
///   @override
///   JsonObject toJson() => {'name': name, 'age': age};
/// }
/// ```
@immutable
class ApiCall<T> {
  /// The [RequestFactory] of this API call
  final RequestFactory request;

  /// The [ResponseConverter] of this API call
  final ResponseConverter<T> convert;

  const ApiCall._(this.request, this.convert);

  /// Create an [ApiCall] that converts json element into [T]
  ApiCall.json({required RequestFactory request, required FromJson<T> fromJson})
      : this._(request, (response) => fromJson(json.decode(response.body)));

  /// Create an [ApiCall] returns the raw body as a String
  static ApiCall<String> raw({required RequestFactory request}) =>
      ApiCall<String>._(request, (response) => response.body);

  /// Create an [ApiCall] that converts [JsonObject] into [T]
  static ApiCall<T> object<T>({required RequestFactory request, required FromJsonObject<T> fromJson}) =>
      ApiCall<T>.json(request: request, fromJson: (body) => fromJson(body));

  /// Create an [ApiCall] that converts json List into a List of [T]
  static ApiCall<List<T>> list<T>({required RequestFactory request, required FromJsonObject<T> fromJson}) =>
      ApiCall<List<T>>.json(
          request: request, fromJson: (body) => (body as List).map((body) => fromJson(body)).toList(growable: false));

  /// Perform the API call to [client] with given [params] and get the result asynchronously.
  Future<T> call(XHttpClient client, [Map<String, dynamic> params = const {}]) => request(client, params).then(convert);

  /// Add this call after your API declaration to mock it by returning a mock [response]
  /// You can also:
  /// - Set a [delay] to simulate server latency
  /// - Make the mock response dynamic by using given parameters
  /// - Throw an [XHTTPException] to simulate errors
  ApiCall<T> mock({Duration delay = const Duration(seconds: 1), required FromJson<T> response}) =>
      _MockApiCall<T>(delay, response, convert);
}

@immutable
class _MockApiCall<T> extends ApiCall<T> {
  final Duration delay;
  final FromJson<T> response;

  const _MockApiCall(this.delay, this.response, ResponseConverter<T> convert) : super._(_request, convert);

  @override
  Future<T> call(XHttpClient client, [Map<String, dynamic> params = const {}]) =>
      Future.delayed(delay).then((_) => response(params));

  static Future<Response> _request(XHttpClient client, Map<String, dynamic> params) => throw UnimplementedError();
}
